package eu.jonahbauer.wizard.core.machine.states.round;

import eu.jonahbauer.wizard.core.machine.Game;
import eu.jonahbauer.wizard.core.machine.GameData;
import eu.jonahbauer.wizard.core.machine.states.trick.StartingTrick;
import eu.jonahbauer.wizard.core.messages.observer.PredictionMessage;
import eu.jonahbauer.wizard.core.messages.observer.UserInputMessage;
import eu.jonahbauer.wizard.core.messages.player.PlayerMessage;
import eu.jonahbauer.wizard.core.messages.player.PredictMessage;
import lombok.Getter;
import org.jetbrains.annotations.VisibleForTesting;

import java.util.Collections;
import java.util.UUID;

import static eu.jonahbauer.wizard.core.messages.observer.UserInputMessage.Action.MAKE_PREDICTION;

@Getter
public final class Predicting extends RoundState {
    private transient Integer prediction;

    public Predicting(Game game, GameData data) {
        super(game, data.requireAllHands().requirePredictions().requireTrumpSuit().requireCurrentPlayer());
    }

    @Override
    public void onEnter() {
        getGame().notify(new UserInputMessage(getCurrentPlayer(), MAKE_PREDICTION, getTimeout(true)));
        timeout();
    }

    @Override
    public void onTimeout() {
        try {
            checkPrediction(0);
            this.prediction = 0;
            transition();
        } catch (IllegalArgumentException e) {
            try {
                checkPrediction(1);
                this.prediction = 1;
                transition();
            } catch (IllegalArgumentException e2) {
                throw new AssertionError(e2);
            }
        }
    }

    @Override
    public void onExit() {
        if (prediction != null) {
            getGame().notify(new PredictionMessage(getCurrentPlayer(), prediction));
        }
    }

    @Override
    public void onMessage(UUID player, PlayerMessage message) {
        if (getCurrentPlayer().equals(player) && message instanceof PredictMessage predictMessage) {
            checkPrediction(predictMessage.getPrediction());
            this.prediction = predictMessage.getPrediction();
            transition();
        } else {
            super.onMessage(player, message);
        }
    }

    @VisibleForTesting
    void checkPrediction(int prediction) {
        if (prediction < 0 || prediction > getRound() + 1) {
            throw new IllegalArgumentException("Prediction must be between 0 and " + (getRound() + 1) + ".");
        }

        if (!getGame().getConfig().allowExactPredictions() && isLastPlayer()) {
            int sum = getPredictions().values().stream().mapToInt(i -> i).sum();
            if (sum + prediction == getRound() + 1) {
                throw new IllegalArgumentException("Predictions must not add up.");
            }
        }
    }

    private void transition() {
        if (prediction == null) throw new AssertionError();

        GameData data = getData().withPrediction(getCurrentPlayer(), prediction)
                                 .withNextPlayer();

        if (isLastPlayer()) {
            transition(new StartingTrick(
                    getGame(),
                    data.withTrick(0).withTricks(Collections.emptyMap())
            ));
        } else {
            transition(new Predicting(
                    getGame(),
                    data
            ));
        }
    }

    private boolean isLastPlayer() {
        return getDealer().equals(getCurrentPlayer());
    }
}
