package eu.jonahbauer.wizard.core.machine.states.trick;

import eu.jonahbauer.wizard.core.machine.Game;
import eu.jonahbauer.wizard.core.machine.GameData;
import eu.jonahbauer.wizard.core.messages.observer.CardMessage;
import eu.jonahbauer.wizard.core.messages.observer.UserInputMessage;
import eu.jonahbauer.wizard.core.messages.player.PlayCardMessage;
import eu.jonahbauer.wizard.core.messages.player.PlayerMessage;
import eu.jonahbauer.wizard.core.model.Card;
import org.jetbrains.annotations.VisibleForTesting;

import java.util.Collection;
import java.util.List;
import java.util.UUID;

import static eu.jonahbauer.wizard.core.messages.observer.UserInputMessage.Action.PLAY_CARD;

public final class PlayingCard extends TrickState {
    private transient Card card;

    public PlayingCard(Game game, GameData data) {
        super(game, data.requireStack());
        assert assertions();
    }

    @Override
    public void onEnter() {
        getGame().notify(new UserInputMessage(getCurrentPlayer(), PLAY_CARD, getTimeout(true)));
        timeout();
    }

    @Override
    public void onMessage(UUID player, PlayerMessage message) {
        if (getCurrentPlayer().equals(player) && message instanceof PlayCardMessage cardMessage) {
            checkCard(cardMessage.getCard());
            this.card = cardMessage.getCard();
            transition();
        } else {
            super.onMessage(player, message);
        }
    }

    @Override
    public void onTimeout() {
        var hand = getHands().get(getCurrentPlayer());
        this.card = hand.stream().filter(c -> {
            try {
                checkCard(c);
                return true;
            } catch (IllegalArgumentException e) {
                return false;
            }
        }).findAny().orElseThrow(() -> new AssertionError("Cannot play any card."));
        transition();
    }

    @Override
    public void onExit() {
        if (card != null) {
            getGame().notify(new CardMessage(getCurrentPlayer(), card));
        }
    }

    @VisibleForTesting
    void checkCard(Card card) {
        var hand = getHands().get(getCurrentPlayer());
        if (!hand.contains(card)) {
            throw new IllegalArgumentException("You do not have this card on your hand.");
        }

        Card.Suit suit = getTrickSuit();
        if (card.getSuit().isColor() && suit != null && card.getSuit() != suit) {
            if (hand.stream().anyMatch(c -> c.getSuit() == suit)) {
                throw new IllegalArgumentException("Must follow suit.");
            }
        }
    }

    private void transition() {
        if (card == null) throw new AssertionError();

        GameData data = getData().withCardPlayed(getCurrentPlayer(), card);

        var summary = data.getHands().values().stream()
                          .mapToInt(Collection::size)
                          .summaryStatistics();

        if (summary.getMax() == summary.getMin()) {
            transition(new FinishingTrick(getGame(), data));
        } else {
            transition(new PlayingCard(getGame(), data.withNextPlayer()));
        }

    }

    private boolean assertions() {
        int expectedSize = getRound() - getTrick();
        for (List<Card> hand : getHands().values()) {
            if (hand.size() != expectedSize && hand.size() != expectedSize + 1) {
                throw new AssertionError("Unexpected hand size: " + hand.size() + " (expected " + expectedSize + " or " + (expectedSize + 1) + ").");
            }
        }
        int size = getHands().get(getCurrentPlayer()).size();
        if (size != expectedSize + 1) {
            throw new AssertionError("Unexpected current player's hand size: " + size + " (expected " + (expectedSize + 1) + ")");
        }
        return true;
    }
}
